# -*- coding: utf-8 -*-
"""
Created on Wed Jun 14 14:49:24 2017

@author: akidribm
"""

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()

# Imports sortal sort types, functions and rules
from sortal import (transformSet, similarityMap, rule, primitiveSort, crossAlgebraPlane2D, point2D, 
                    coordinate, lineSegment2D, line2D, label, weight, disjunctiveSort, attributeSort, 
                    affineTransform,  vector2D, planeSegment2D)
from sortal import description, primitiveSort, disjunctiveSort, rule, sdlParser, numeric, parseReader

# Imports GUI Library
from pylab import *
import matplotlib
import matplotlib.gridspec as gridspec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import  matplotlib.pylab as plt
import matplotlib.pyplot as pltplot
from matplotlib.collections import LineCollection, PatchCollection
matplotlib.use ("gTkAgg")
import tkinter as tk
from tkinter import ttk, Tk, Label, Button, Frame
from tkinter import *
from matplotlib.widgets import RectangleSelector
from copy import deepcopy
import random


if __name__ == "__main__":
    
    LARGE_FONT= ("Verdana", 16)
    NORM_FONT= ("Verdana", 11)

    rule.cleanup()
    sdlParser.parse('sdl_files/parametric_ice_ray.sdl')

    r0 = sdlParser.rules['r_zero']
    r1 = sdlParser.rules['r_triangle_quad']
    r2 = sdlParser.rules['r_quad_penta_triangle']
    r3 = sdlParser.rules['r_two_quads']
    r4 = sdlParser.rules['r_penta_quad']
    formM1 = sdlParser.forms['test0a1']
    rules = [r0, r1, r2, r3, r4]
        
    class Application(Frame):
        print ("<<<< GENERATING APPLICATION WINDOW >>>>")
        fig_main = Figure(figsize = (3, 5), dpi = 150) # Figure contains 'Current Drawing'
        fig_options = Figure(figsize = (11, 8), dpi = 90) # Figure contains 'Matches, Results'
        fig_rules = Figure(figsize = (2, 2), dpi = 100) # Figure contains 'Rules'

        # Sets Current Drawing to formL4
        # Current Drawing can be changed by changing index of form_set
        form_main = deepcopy(formM1)
        form_test = deepcopy(formM1) 
  
        matchForms = [] # Holds list of match instances for all three rules
        resultForms = [] # Holds list of result instances for chosen match
        event_list = ['o'] # Holds list of events in the order they occur in, i.e. o - Starting Window, m - Matches, r - Results
        get_ax_index = [] # Holds list of figure subplot instances in fig_options
        rules_sides = [] # Holds list of LHS and RHS of each rule   
        back = []
        back_main = []
        back_option_chosen = []
        new = False

        """
        ================================
        APPLICATION'S MAIN FUNCTION LOOP
        ================================
        """
        def __init__(self, master = None):
            Frame.__init__(self, master)
            self.grid()
            self.master.title("SGI Test Case: Parametric Ice Ray Grammar (No Labels)")
    
            # Generates tkinter window grid with 6 rows and 5 columns
            for r in range(6):
                self.master.rowconfigure(r, weight=1)    
            for c in range(5):
                self.master.columnconfigure(c, weight=1)

            # Holds 'Current Drawing' figure (fig_main)
            self.Frame1 = Frame(master, bg="gray")
            self.Frame1.grid(row = 0, column = 0, rowspan = 6, columnspan = 1, sticky = W+E+N+S) 
            self.frame1_text = "Initial Shape"
            self.Frame1_label = Label (master = self.Frame1, text = self.frame1_text, background = "gray", font = LARGE_FONT, anchor = "center")
            self.Frame1_label.pack()
            self.Frame1a_label = Label (master = self.Frame1, text = "Use mouse cursor to select area \n of interest in Current Shape plot. \nPink: defines a plane\nA: quadrilaterals\nB: pentagons\nC: triangles", background = "#3ECE90", font = NORM_FONT, anchor = "center", padx = 100, pady = 10, wraplength = 800)
            self.Frame1a_label.pack()
            
            # Holds 'Matches/Results' figure (fig_options)
            self.Frame2 = Frame(master, bg='#3ECE90')
            self.Frame2.grid(row = 0, column = 1, rowspan = 6, columnspan = 3, sticky = W+E+N+S)
            self.Frame2_label = Label (master = self.Frame2, text = "", background = '#3ECE90', font = LARGE_FONT, anchor = "center")
            self.Frame2_label.pack()
            
            # Holds 'Rules' figure (fig_rules)
            # LHS and RHS figures are side by side for each rule
            self.Frame3 = Frame(master, bg='gray')
            self.Frame3.grid(row = 0, column = 4, rowspan = 6, columnspan = 2, sticky = W+E+N+S)
            self.Frame3_label = Label (master = self.Frame3, text = "Rules", background = 'gray', font = LARGE_FONT, anchor = "center")
            self.Frame3_label.pack()

            # Initalizes option_chosen variable, which takes input from user through mouse clicks
            # This serves as user input on which match to generate results for, which result to apply on the form
            self.option_chosen = IntVar()
            self.matchTest = self.form_main
            self.refresh_matchForms() # Generates initial matches 
            self.menu() # Adds menu 
            self.matchForms_viz() # Draws matches onto Frame 2
            self.select_figure(self.option_chosen) # Select_figure starts visualization of fig_main and fig_options
            
            self.connect() # Connects GUI to mouseclick and hover events 
 
        """
        ================================
        API: MATCH AND RESULT GENERATION
        ================================
        """
        # Refreshes matchForms and resultForms lists after one run of selection, i.e. o -> m -> r or when area is selected
        def refresh_matchForms(self):
            for i in rules:
                i.detect(self.matchTest)
                
        # Generates matches based on which form is assigned to self.form_test
        # The contents of self.matchForms look like [[matches from Rule 1], [matches from Rule 2], [matches from Rule 3]]
        def matchForms_generate(self):
            self.matchForms = []
            self.refresh_matchForms()
            for i in rules:
                self.matchForms.append(i.detection.matches())

        # Generates results based on chosen match from option_chosen variable
        def resultForms_generate(self, option_chosen):
            self.resultForms = [] # Stores result forms 

            self.matchF = self.matchForms_list[option_chosen - 1] # Stores chosen match from list of match forms based on user input
            self.matchRule =  self.rules_log[option_chosen - 1] - 1 # Stores which rule was applied to generate the chosen match stored in self.matchF
            self.resultForms.append(rules[self.matchRule].detection.results(self.matchF)) # Stores result forms generated from chosen match self.matchF with corresponding rule

        # Applies rule to form_test based on chosen match and chosen result
        # rules[self.matchRule].detection.applications contains the results based on the match and the rule
        def confirmForm (self, option_chosen):
            for i in rules[self.matchRule].detection.applications:
                if i.resultForm == self.resultForms_list[option_chosen-1] and i.matchForm == self.matchF: # Checks which rule result corresponds to the chosen result generated by the match forms
                    i.perform(self.form_main)
                    
        """
        ========================
        GRAPHICAL USER INTERFACE
        ========================
        ==========
        EVENT TYPE
        ==========
        Event types refer to the phase the GUI is in.
        
        'o' : Match selection
        
        'm' : Result selection based on match choice
        
        'r' : Result has been selected; form has been transformed/rule has been applied 
              based on result choice, continue selecting matches
              
        One selection run consists of o -> m -> r.
        If an error is triggered (e.g. when the mouse click is outside the subplot box), 
        then the current event type is appended again to event_list.
        """
        
        # Changes Frame 2 Label based on event type
        def text_configure(self, event_type, option_chosen):
            # If first iteration ('o') or results have been chosen ('m' preceeded by 'r'), label shows the number of matches
                
            if event_type == 'o' or event_type == 'r' or event_type == 's': 
                self.Frame2_label.configure(text = "Matches: {}".format(len(self.matchForms_list)))
            elif event_type == 'm': # If match has been selected, shows number of results based on match selected and corresponding rule
                if len(self.event_list) == 3:
                    self.Frame1_label.configure(text = "Current Shape")
                self.Frame2_label.configure(text = "Results: {}".format(len(self.resultForms_list)) + ", Applying Rule {}".format(rules[self.rules_log[option_chosen-1]-1].name))             
                
        # Changes options based on event type to Matches/Results
        def visualization(self, event_type, option_chosen): 
            if event_type == 'o':
                if self.new == True:
                    self.fig_main.clf()
                    self.viz_main(self.form_main, self.fig_main, self.Frame1, create_plot = 'n')
                    self.fig_main.canvas.draw_idle()
                    
                    self.fig_options.clf()
                    self.matchForms_viz()
                    self.fig_options.canvas.draw_idle()
                else:
                    self.viz_main(self.form_test, self.fig_main, self.Frame1, create_plot = 'y') # Draws 'Current Drawing'
                    self.rules_viz() # Draws 'Rules' plots
                    self.matchForms_viz() # Draws 'Matches' plots
                    self.back_main.append(self.form_main)
            elif event_type == 'm':
                self.back_option_chosen.append(option_chosen)
                self.back_main.append(self.form_main)
                self.resultForms_viz(option_chosen) # Draws 'Results' plots and redraws 'Current Drawing' by superimposing chosen match
            elif event_type == 'r':
                self.back_option_chosen.append(option_chosen)
                self.back_main.append(self.form_main)
                self.confirmForm_viz(option_chosen) #  # Applies chosen result's rule to form_test and redraws 'Current Drawing'
                self.matchForms_viz() # Generates new matches based on form_test after rule has been applied and updates 'Matches/Results' frame
            elif event_type == 's':
                self.fig_main.clf()
                self.viz_main(self.form_test, self.fig_main, self.Frame1, create_plot = 'n')
                self.fig_main.canvas.draw_idle()
                
                self.fig_options.clf()
                self.matchForms_viz()
                self.fig_options.canvas.draw_idle()
                
        # Calls visualization and text_configure functions to execute event based on self.option_chosen input
        # Adds next event to self.event_list queue if no error is triggered
        # Calls visualization and text_configure functions to execute event based on self.option_chosen input
        # Adds next event to self.event_list queue if no error is triggered
        def select_figure(self, option_chosen): 
            try:
                event_type = self.event_list[-1]  
                self.visualization(event_type, option_chosen)
                self.text_configure(event_type, option_chosen)
                
                self.toggle_selector = RectangleSelector(self.a_main, self.line_select_callback,
                               drawtype='box', useblit=True,
                               button=[1, 3],  # don't use middle button
                               minspanx=3, minspany=3,
                               spancoords='pixels',
                               interactive=True)
                
                if event_type == 'o':
                    self.event_list.append('m')
                elif event_type == 'm':
                    self.event_list.append('r')
                elif event_type == 'r':
                    self.event_list.append('m')
                elif event_type == 's':
                    self.event_list.append('m')

            except:
                self.error_box()
                self.event_list.append(event_type)

                
        # Visualizes rules' LHS and RHS on rightmost part of GUI window
        def rules_viz(self):
            if self.new == False:
                self.rules_sides = []
                for i in rules:
                    self.rules_sides.append(i.lhs)
                    self.rules_sides.append(i.rhs)
                self.viz_options(self.rules_sides, self.fig_rules, self.Frame3, create_plot = 'y', create_label ='ru', col_count = 2, onclick_option = 'n')

        # Initializes menu
        def menu(self):
            menubar = Menu (root)  # Saves PDF of 'Current Drawing' only
            menubar.add_command(label="Start Over", command=self.start_over) # Closes window
            menubar.add_command(label="Start from Current Form", command=self.start_over_c) # Go back to previous form
            menubar.add_command(label="Save Current Drawing", command=self.save) # Saves drawing as pdf
            menubar.add_command(label="Go Back A Selection", command=self.selection) # Closes window
            menubar.add_command(label="Exit", command=root.destroy)
            root.config(menu=menubar)

        # When Start Over button is selected on menu
        # GUI starts from original form
        def start_over(self):
            self.event_list.append('s')
            self.form_test = deepcopy(formM1)
            self.form_main = deepcopy(formM1)
            self.select_figure(self.option_chosen)
            
        # Starts over from 'current drawing' form 
        def start_over_c(self):
            self.event_list.append('s')
            self.form_test = self.form_main
            self.select_figure(self.option_chosen)
            
        #  Goes back one selection step until 'current drawing' form
        def selection(self):
            if len(self.back) != 0:
                self.event_list.append('s')
                self.form_test = self.back[-1]
                self.back.pop(-1)
                self.select_figure(self.option_chosen)
            else:
                error_box = Tk()
                error_box.title("Input Error")
                Label(error_box, text = "There are no previous selections to go back to.", padx = 5, pady = 15).pack()
                button = Button(error_box, text= "OK", padx = 10, pady = 5, command= error_box.destroy)
                button.pack()
                error_box.update()
                self.errorBox = error_box
       
        # Saves file of 'Current Drawing'
        # File type may be changed by change the .pdf extension to e.g.: .jpg, .png 
        def save(self):
            self.fig_main.savefig('parametric_selection.pdf', bbox_inches='tight')
            
        # Identifies which subplot the mouseclick was in, calls self.select_figure() to execute next event
        def on_press(self, event):
            if event.inaxes in self.get_ax_index and event.xdata != None and event.ydata != None:
                for i, ax in enumerate(self.get_ax_index): 
                   if ax == event.inaxes:
                        self.option_chosen = i+1
                self.select_figure(self.option_chosen)

        # Identifies which subplot the mouse is hovering over, changes bg color of chosen subplot
        def on_motion(self, event):
            for ax in self.get_ax_index: 
               if ax == event.inaxes:
                   ax.set_axis_bgcolor('0.55')  
            self.fig_options.canvas.draw_idle()

        # Reverses bg color of subplot to white after mouse cursor leaves the subplot
        def off_motion(self, event):
            for ax in self.get_ax_index: 
                   if ax == event.inaxes:
                       ax.set_axis_bgcolor("white")
            self.fig_options.canvas.draw_idle()    
            
        # Takes mouseclick data and turns it into boundaries of selection plane
        # Stores previous 'Current Drawing' instance
        # Performs cross algebra on form_main and returns it to form_test
        # Appends event to event_list
        # Calls main event function (select_figure) and updates plots
        def line_select_callback(self, eclick, erelease):
            'eclick and erelease are the press and release events'
            x1, y1 = eclick.xdata, eclick.ydata
            x2, y2 = erelease.xdata, erelease.ydata

            v1 = vector2D(x1, y1)
            v2 = vector2D(x1, y2)
            v3 = vector2D(x2, y2)
            v4 = vector2D(x2, y1)
            self.bound = [v1, v2, v3, v4]
            
            if len(self.back) == 0:
                self.back.append(self.form_main)
            
            if (self.event_list[-2] == 's' and self.event_list[-1]=='m'): 
                self.back.append(self.form_test)
            elif self.event_list[-2] == 'r' and self.event_list[-1] == 's':
                self.back.append(self.form_main)
                
            self.form_test = crossAlgebraPlane2D.freeSelection(self.bound, self.form_main)
            self.event_list.append('s')
            self.select_figure(self.option_chosen)
            
        # Connects mouse event functions to the 'Matches/Results' figure (fig_options)
        def connect(self):
            self.cidpress = self.fig_options.canvas.mpl_connect('button_press_event', self.on_press)
            self.cidmotion_on = self.fig_options.canvas.mpl_connect('axes_enter_event', self.on_motion)
            self.cidmotion_off = self.fig_options.canvas.mpl_connect('axes_leave_event', self.off_motion)
            self.rspress = self.fig_main.canvas.mpl_connect('key_press_event', self.toggle_selector)

        # Resets entire GUI to original form
        def resetY(self):
            self.errorBox.destroy()
            self.event_list.append('s')
            self.form_test = self.form_main
            self.select_figure(self.option_chosen)
            
        # Sets form_test to previous drawing or back to form_main.
        # If there are no other matches then, it will raise an error.
        def resetB(self):
            self.errorBox.destroy()
            self.event_list.append('s')
            self.form_test = self.back[-1]
            self.back.pop(-1)
            self.select_figure(self.option_chosen)
            
        # Closes error box and tkinter window when 'Exit' button is pressed.
        def end_function(self):
            self.errorBox.destroy()
            root.destroy()
            
        # Displays Error Dialog Box when the select_figure function gets an error,
        # when the mouse click is outside the fig_options subplot areas
        # or when there are no matches found in the selection area
        def error_box(self):
            error_box = Tk()
            error_box.title("Input Error")
            buttonList = []
            
            if self.event_list[-1] == 'm':
                message = "Please click directly on the plot of the match."
                Label(error_box, text = message, padx = 5, pady = 15).pack()
                button = Button(error_box, text= "OK", padx = 10, pady = 5, command= error_box.destroy)
                buttonList.append(button)
                
            elif self.event_list[-1] == 'r':
                if len(self.matchForms_list) == 0:
                    message = "No matches have been found."
                else:
                    message = "Please click directly on the plot of the result."
                Label(error_box, text = message, padx = 5, pady = 15).pack()
                button = Button(error_box, text= "OK", padx = 10, pady = 5, command= error_box.destroy)
                buttonList.append(button)
                
            elif self.event_list[-1] == 's' and (len(self.matchForms_list) == 0 or len(self.resultForms_list == 0)):
                if len(self.matchForms_list) == 0 :
                    label_text = "No matches found.\nWould you like to start again or go back?"
                elif len(self.resultForms_list) == 0:
                    label_text = "No results found. Would you like to start again or go back?"
                    
                Label(error_box, text = label_text, padx = 5, pady = 15).pack()
                yes = Button(error_box, text= "Start from Current From", padx = 10, pady = 5, command = self.resetY)
                back = Button(error_box, text= "Go Back", padx = 10, pady = 5, command = self.resetB)
                no = Button(error_box, text= "Exit", padx = 10, pady = 5, command = self.end_function)
                buttonList = [yes, back, no]
                
            for i in buttonList:
                i.pack(side = TOP)  
                
            error_box.update()
            self.errorBox = error_box
          
        # Generates and plots matchForms in 'Matches/Results' frame (Frame 2)            
        def matchForms_viz(self):
            self.rules_log = []
            self.matchForms_list = []

            # Changes which form to generate matches for
            # If most recent event was selection ('s') or starting ('o') event, 
            # then the form to generate matches for is the selected area (self.form_test)
            # If most recent event was result selection ('result') or match selection ('m')
            # then the form to generate matches for is the main form (self.form_main)
            if self.event_list[-1] == 's' or self.event_list[-1] == 'o':
                self.matchTest =  self.form_test
            elif self.event_list[-1] == 'r' or self.event_list[-1] == 'm':
                self.matchTest = self.form_main
                
            self.matchForms_generate()
            self.super_forms = [] 

            # Turns self.matchForms into a single-layered list from the list of lists and uses 
            # the index of the list inside the list as the rule number
            # Generates copies of form_test to superimpose matchForms on
            for a, i in enumerate(self.matchForms):
                for j in i:
                    self.matchForms_list.append(j)
                    self.super_forms.append(self.matchTest)
                    self.rules_log.append(a+1)
                    
            if len(self.matchForms_list) == 0:
                self.error_box()
                
            # Updates 'Matches/Results' figure by drawing match options superimposed on form_test
            self.fig_options.clf()
            if len(self.event_list) == 1:
                self.viz_options(self.super_forms, self.fig_options, self.Frame2, create_plot = 'y', super_color = '#3ECE90')
                self.event_list.append('o')
            else:
                self.viz_options(self.super_forms, self.fig_options, self.Frame2, super_color = '#3ECE90')

            self.viz_options(self.matchForms_list, self.fig_options, self.Frame2, super_color = "blue", create_label = "m")
            self.fig_options.canvas.draw_idle()
                    
        # Generates and plots matchForms in 'Matches/Results' frame (Frame 2) 
        def resultForms_viz(self, option_chosen):
            self.resultForms_generate(option_chosen) # Generates results based on chosen match 
            self.resultForms_list = []
            super_forms = [] 
            matchForm_sup = [] 
            for i in self.resultForms:
                for j in i:
                    self.resultForms_list.append(j)
                    super_forms.append(self.form_main) # For superimposing result figures onto form_test
                    matchForm_sup.append(self.matchF) # For superimposing chosen match figures onto form _test

            # Updates 'Current Drawing' with chosen match superimposed on form_test figure
            self.fig_main.clf()
            self.viz_main(self.form_main, self.fig_main, self.Frame1)
            self.viz_main(self.matchF, self.fig_main, self.Frame1, super_color = "#3ECE90")
            self.fig_main.canvas.draw_idle()
            
            # Updates 'Matches/Results' with results generated from chosen match
            # Draws both match and result (superimposed) on form
            self.fig_options.clf()
            self.viz_options(super_forms, self.fig_options, self.Frame2, super_color = '#3ECE90')
            self.viz_options(matchForm_sup, self.fig_options, self.Frame2, super_color = "blue")
            self.viz_options(self.resultForms_list, self.fig_options, self.Frame2, super_color = '#FCA309', create_label = "r")
            self.fig_options.canvas.draw_idle()
        
        # Applies rule of chosen result to form_test and redraws 'Current Drawing' after rule has been applied
        def confirmForm_viz(self, option_chosen):
            self.back_main.append(self.form_main)
            self.confirmForm(option_chosen)
            
            # Updates 'Current Drawing' with form_test after rule application   
            self.fig_main.clf()
            self.viz_main(self.form_main, self.fig_main, self.Frame1)
            self.fig_main.canvas.draw_idle()
            
        # Prepares plane vertices for plotting as patches; appends each vector coordinate pair to vertex_list  
        def data_prep_plane(self, vertex_source):
            vertex_list = []
            attributeValue = []

            if vertex_source.hasAttribute():
                for i in vertex_source.attribute.elements:
                
                    attributeValueTemp = self.parseText(i.elements[0].getValue())

                    if len(attributeValueTemp) == 1:
                        attributeValue.append(attributeValueTemp[0])
                    elif len(attributeValueTemp) == 3:
                        try:
                            r = int(attributeValueTemp[0])
                            g = int(attributeValueTemp[1])
                            b = int(attributeValueTemp[2])
                            attributeValue.append('#%02x%02x%02x' % (r, g, b))
                        except:
                            tempText = ''
                            for thing in attributeValueTemp:
                                tempText += thing
                                
                            attributeValue.append(tempText)
                            
            for j in vertex_source.boundaries[0].polygonal:
                vertex_list.append([j.x, j.y])

            return vertex_list, attributeValue
            
        # Extracts and prepares line coordinates of each line segment in form for visualization 
        # For this case, a line segment is two sets of vertices connected to each other, so plotting is 
        # from [x1, y1] to [x2, y2], from [x3, y3] to [x4, y4] and so on
        def data_prep_line(self, line_set, w, h):
            # w - number of coordinates, e.g. (x, y) -> w = 2
            # h - number of lines in form
            lineSeg_x = [[0 for x in range(w)] for y in range(h)]
            lineSeg_y = [[0 for x in range(w)] for y in range(h)]
            attribute_set = []
            
            for i in range (h):

                if line_set[i].hasAttribute():
                    tempList = []

                    for obj in line_set[i].attribute.elements:
                        tempList.append(self.parseText(obj.elements[0].getValue()))

                    attribute_set.append(tempList)
                else:
                    attribute_set.append('')

                for j in range (w):
                    if j == 0:  # For x,y and potentially z coordinates
                        if line_set[i].start.type == 'struct-vector2Dfloat':
                            lineSeg_x[i][j] = float(round(line_set[i].start.x, 2))
                            lineSeg_y[i][j] = float(round(line_set[i].start.y, 2))
                        elif line_set[i].start.type == 'struct-vector2D':
                            lineSeg_x[i][j] = float(line_set[i].start.x.value)
                            lineSeg_y[i][j] = float(line_set[i].start.y.value)
                    else:
                        if line_set[i].end.type == 'struct-vector2Dfloat':
                            lineSeg_x[i][j] = line_set[i].end.x
                            lineSeg_y[i][j] = line_set[i].end.y
                        elif line_set[i].start.type == 'struct-vector2D':
                            lineSeg_x[i][j] = line_set[i].end.x.value
                            lineSeg_y[i][j] = line_set[i].end.y.value

            return lineSeg_x, lineSeg_y, attribute_set
        
        # Converts description contents to strings without quotes or double quotes
        def parseText(self, a):
            #if a[1] == '(':
            reader = parseReader(a)
            
            values = []
            for i in range(len(a)):
                
                if reader.token() == '`' or reader.token() == '(' or reader.token() == ',':
                    pass
                elif reader.token() == '"':
                    value = reader.value()[1:-1]
                    values.append(value)
                elif reader.token() == 'a':
                    value = reader.value()
                    values.append(value)
                elif reader.token() == '0':
                    value = reader.value()
                    values.append(value)
                elif reader.token() == '+':
                    value = reader.value()
                    values.append(value)
                elif reader.token() == ')':
                    break
                reader.newToken()
            
            return values
            #else:
                #return

        # Updates 'Current Drawing' frame (Frame 1)
        def viz_main(self, form, figure_House, Frame_house, super_color = 'black', super_face_color = '#FFB6C1', create_plot = 'n', superform ='y'):
            self.a_main = self.fig_main.add_subplot(111)
            self.fig_main.subplots_adjust(left=0.05, bottom=0.5)
            
            # Formats 'Current Drawing' Figure
            self.a_main.axes.set_anchor('C')   
            self.a_main.margins(y=0.1)
            self.a_main.margins(x=0.1)
            self.a_main.axes.get_xaxis().set_visible(False)
            self.a_main.axes.get_yaxis().set_visible(False)
            self.a_main.axes.set_aspect(1.0)
            self.a_main.set_xlim(-2, 9)
            self.a_main.set_ylim(-2.5, 11)

            complete_desc1 = [] 
            complete_desc2 =[]
            complete_desc3 = []
            count = 0
            
            for obj in form.elements:
                cell_text = []
                
                if obj.type == 'form-interval':
                    elements = obj.elements
                    lineSeg_x, lineSeg_y, attribute_set = self.data_prep_line(elements, 2, h = len (obj.elements))

                    for j in range(len(lineSeg_x)):
                        
                        lineColor = super_color
                        if len(attribute_set[j])!=0:
                            for attrib in attribute_set[j]:
                                if type(attrib) == str or type(attrib[0]) == str:
                                    coords = [(lineSeg_x[j][0] + lineSeg_x[j][1])/2, (lineSeg_y[j][0] + lineSeg_y[j][1])/2]
                                              
                                    #self.a_main.annotate(attrib[0], xy = (coords[0], coords[1]), xycoords='data',  xytext=(0, 0), size = 5, textcoords='offset points', ha = 'right', va = 'bottom', bbox = dict(boxstyle = 'round,pad=0.2', fc = 'white'))

                        # Plots line segment in subplot
                        self.a_main.plot(lineSeg_x[j], lineSeg_y[j], color = lineColor, linewidth = 1.3, picker = 2)
                       
                elif obj.type == 'form-discrete':
                    if 'ind-description' in obj.sort.baseSort().characteristic.type:
                        count+=1
                        for a in obj.elements:
                            
                            if len(obj.elements) >1:
                                
                                values = self.parseText(a.getValue())
                                values.insert(0, a.sort.name + ' ' + str(obj.elements.index(a)+1))
                            else:
                                values = self.parseText(a.getValue())
                                values.insert(0, a.sort.name)

                            if len (values) == 4:
                                complete_desc1.append(values)
                            elif len(values) == 3:
                                complete_desc2.append(values)
                            elif len(values) == 2:
                                complete_desc3.append(values)
                            
                    elif 'ind-point2D' in obj.sort.baseSort().characteristic.type:
                        for a in obj.elements:
                            if a.hasAttribute():
                              attributes = a.attribute.elements
                              value_attributes = []
                              for attrib in attributes:

                                  if attrib.sort.name == 'desc':
                                      text= self.parseText(attrib.elements[0].getValue())
                                    
                                  elif attrib.sort.name == 'plbl':
                                      text= self.parseText(attrib.getValue())
                                      
                                  elif attrib.sort.name == 'label': 
                                      text = attrib.elements[0].getValue()
                                      if superform == 'y':
                                          super_color = 'pink'
                                        
                                  #self.a_main.annotate(text[0], xy = (a.value.x, a.value.y), xycoords='data', xytext=(-2.0, -2.0), size = 6, textcoords='offset points', ha = 'right', va = 'bottom')# bbox = dict(boxstyle = 'round,pad=0.2', fc = super_face_color))
                                 
                            else:
                                self.a_main.plot(a.value.x, a.value.y, 'wo', markersize = 5, markeredgecolor = super_face_color, markerfacecolor = super_color, markeredgewidth=0.7) 

                            
                elif obj.type == 'form-coForm':
                    elements = obj.elements
                    for element in elements:
                        pts, attributeValue = self.data_prep_plane(element)
                        edge_color = super_face_color

                        for index, attrib in enumerate(attributeValue):
                            """   
                            if 'A' in attrib or 'C' in attrib or 'B' in attrib:
                                self.a_main.annotate(attrib, xy = (pts[index][0]+1, pts[index][1]), xycoords='data', xytext=(-2, -2), size = 3.7, textcoords='offset points', ha = 'right', va = 'bottom', bbox = dict(boxstyle = 'round,pad=0.2', fc = 'white')) 
                                self.a_main.plot(pts[0][0], pts[0][1], 'wo', markersize = 3, markeredgecolor = super_color, markerfacecolor = super_color, markeredgewidth=0.7) 
                            
                            
                            if attrib[0] == '#':
                                edge_color = attrib
                            else:
                                r = lambda: random.randint(100,255)
                                g = lambda: random.randint(0,150)
                                b = lambda: random.randint(0,100)
                                edge_color = ('#%02X%02X%02X' % (r(),g(),b())) 
                            """ 
                        patch = plt.Polygon(pts, facecolor = edge_color, edgecolor = 'grey')
                        self.a_main.add_patch(patch)
    
            if superform == 'y':  
                if len(complete_desc1) != 0:
                    the_table1 = self.a_main.table(cellText=complete_desc1, colLoc='upper left', loc='top', bbox =[-0.9, -1.1, 2.0, 1.5])
                    the_table1.set_fontsize(5)
                    table_props1 = the_table1.properties()
                    table_cells1 = table_props1['children']
                    for cell in table_cells1: cell.set_height(0.11)
                    
                if len(complete_desc2) != 0:
                    the_table2 = self.a_main.table(cellText=complete_desc2, colLoc='upper left', loc='top',  bbox =[0, -1.1, 1.75, 0.7])
                    the_table2.set_fontsize(5)
                    table_props2 = the_table2.properties()
                    table_cells2 = table_props2['children']
                    for cell in table_cells2: cell.set_height(0.11)

                if len(complete_desc3) != 0:
                    the_table3 = self.a_main.table(cellText=complete_desc3, colLoc='upper right', loc='bottom')
                    the_table3.set_fontsize(5)
                    table_props3 = the_table3.properties()
                    table_cells3 = table_props3['child_artists']
                    for cell in table_cells3: cell.set_height(0.1)
  
            # Creates entirely new figure
            if create_plot == 'y' and self.new == False:
                self.canvas_main = FigureCanvasTkAgg(figure_House, Frame_house)
                self.canvas_main.get_tk_widget().pack(fill = BOTH, expand = 1) 

                self.fig_main.set_canvas(self.canvas_main)
                # Adding prev,next button
                axprev = self.fig_options.add_subplot(2,1,2)
                axnext = self.fig_options.add_subplot(2,1,2)
                
        # Updates 'Matches/Results' frame (Frame 2) 
        def viz_options(self, form_list, figure_House, Frame_house, create_plot = 'n', super_color = 'black', super_face_color = 'pink', superform = 'n', create_label =  "n", col_count = 3, onclick_option = 'y'):
            # Refreshes index counter, list of subplot instances, and adjusts column count based on number of forms to be visualized
            self.get_ax_index = []
            if len(form_list) <2:
                col_count = 1
                textsize = 12
            elif len(form_list) == 2 or len(form_list) <= 12:
                col_count = 2
                textsize = 10
            elif len(form_list) > 12 and len(form_list) <=27:
                col_count = 3
                textsize = 8
            elif len(form_list) >27:
                col_count = 4
                textsize = 7
             
            ind = 1 # Keeps track of subplot number
            
            # Plots Matches/Results in Frame 2
            for counter,form in enumerate(form_list):
                row_count = ceil(len(form_list)/col_count)
                
                self.a_options = figure_House.add_subplot(row_count, col_count, ind)

                # Adjusts ratio of height and width of each subplot in 'Matches/Results' depending on number of subplots generated.
                self.a_options.axes.set_aspect(1.0)

                if create_label == "m":
                    self.a_options.set_title('%d'% (ind,)+': Rule {}'.format(self.rules_log[counter])+'\n', fontsize = textsize)

                elif create_label == "r":
                    self.a_options.set_title('Result {}'.format(ind), fontsize = textsize)
                    
                elif create_label == "ru":
                    self.a_options.set_xlim(-7,7)
                    self.a_options.set_ylim(-2.5,8)
                    extra_label=''
                    
                    # Sets text inside title for 'Rules' subplots
                    counter = ind%2 + math.floor(ind/2)
                        
                    if ind%2 == 0:
                        extra_label = "RHS\n"
                    else:
                        extra_label = "LHS\n"

                    self.a_options.set_title(extra_label + 'Rule {}'.format(counter), fontsize = 11)   
                    
                """ PLOTTING OPTIONS """
                cell_text = []
                complete_desc1 = []
                complete_desc2 = []
                complete_desc3 = []
                count = 0 

                for obj in form.elements:
                    if obj.type == 'form-interval':
                        elements = obj.elements
                        lineSeg_x, lineSeg_y, attribute_set = self.data_prep_line(elements, 2, h = len (obj.elements))
    
                        for j in range(len(lineSeg_x)):
                            
                            lineColor = super_color
                            if len(attribute_set[j])!=0:
                                for attrib in attribute_set[j]:
                                    if type(attrib) == str or type(attrib[0]) == str:
                                        coords = [(lineSeg_x[j][0] + lineSeg_x[j][1])/2, (lineSeg_y[j][0] + lineSeg_y[j][1])/2]
                                                  
                                        #self.a_options.annotate(attrib[0], xy = (coords[0], coords[1]), xycoords='data',  xytext=(0, 0), size = 7, textcoords='offset points', ha = 'right', va = 'bottom')# bbox = dict(boxstyle = 'round,pad=0.2', fc = 'white'))
    
                            # Plots line segment in subplot
                            self.a_options.plot(lineSeg_x[j], lineSeg_y[j], color = lineColor, linewidth = 1.3, picker = 2)
                           
                    elif obj.type == 'form-discrete':
                        if 'ind-description' in obj.sort.baseSort().characteristic.type:
                            count+=1
                            for a in obj.elements:
                                if len(obj.elements) >1:
                                    values = self.parseText(a.getValue())
                                    values.insert(0, a.sort.name + ' ' + str(obj.elements.index(a)+1))
                                else:
                                    values = self.parseText(a.getValue())
                                    if values != None:
                                        values.insert(0, a.sort.name)

                            if values != None:       
                                if len (values) == 4:
                                    complete_desc1.append(values)
                                elif len(values) == 3:
                                    complete_desc2.append(values)
                                elif len(values) == 2:
                                    complete_desc3.append(values)

                        elif 'ind-point2D' in obj.sort.baseSort().characteristic.type:
                            for a in obj.elements:
                                if a.hasAttribute():
                                  attributes = a.attribute.elements
                                  value_attributes = []
                                  for attrib in attributes:
                                      if attrib.sort.name == 'desc':
                                          text= self.parseText(attrib.elements[0].getValue())
                                    
                                      elif attrib.sort.name == 'plbl':
                                          text= self.parseText(attrib.getValue())
                                      
                                      elif attrib.sort.name == 'label': 
                                          text = attrib.elements[0].getValue()
                                          if superform == 'y':
                                              super_color = 'pink'
                                            
                                      self.a_options.annotate(text[0], xy = (a.value.x, a.value.y), xycoords='data', xytext=(0.0, 0.0), size = 7, textcoords='offset points', ha = 'right', va = 'bottom')# bbox = dict(boxstyle = 'round,pad=0.2', fc = 'white'))
                                     
                                else:
                                    self.a_options.plot(a.value.x, a.value.y, 'wo', markersize = 5, markeredgecolor = super_color, markerfacecolor = super_color, markeredgewidth=0.7) 
                        
                    elif obj.type == 'form-coForm':
                        elements = obj.elements
                        for element in elements:
                            pts, attributeValue = self.data_prep_plane(element)

                            edge_color = super_face_color
                            """
                            for index, attrib in enumerate(attributeValue):
                                
                                if ('A' in attrib or 'C' in attrib or 'B' in attrib) and (create_label == 'm' or create_label == 'r'):
                                    self.a_options.annotate(attrib, xy = (pts[index][0]+1, pts[index][1]), xycoords='data', xytext=(-4, -7), size = 7, textcoords='offset points', ha = 'right', va = 'bottom', bbox = dict(boxstyle = 'round,pad=0.2', fc = 'white')) 
                                    self.a_options.plot(pts[index][0], pts[index][1], 'wo', markersize = 3, markeredgecolor = super_color, markerfacecolor = super_color, markeredgewidth=0.7) 
                                
                                if attrib[0] == '#':
                                    edge_color = attrib
                                else:
                                    r = lambda: random.randint(0,255)
                                    edge_color = ('#%02X%02X%02X' % (r(),r(),r())) 
                            """                                    
                            patch = plt.Polygon(pts, facecolor = edge_color, edgecolor = 'grey')
                            self.a_options.add_patch(patch)

                    if superform == 'y':   
                        if len(complete_desc1) != 0:
                            the_table1 = self.a_options.table(cellText=complete_desc1, colLoc='upper right', loc='bottom')
                            table_props1 = the_table1.properties()
                            the_table1.set_fontsize(10)
                            table_cells1 = table_props1['child_artists']
                            for cell in table_cells1: cell.set_height(0.1)
                
                        if len(complete_desc2) != 0:
                            the_table2 = self.a_options.table(cellText=complete_desc2, colLoc='upper right', loc='bottom')
                            table_props2 = the_table2.properties()
                            the_table2.set_fontsize(10)
                            table_cells2 = table_props2['child_artists']
                            for cell in table_cells2: cell.set_height(0.1)
                            
                        if len(complete_desc3) != 0:
                            the_table2 = self.a_options.table(cellText=complete_desc3, colLoc='upper right', loc='bottom')
                            table_props2 = the_table2.properties()
                            the_table2.set_fontsize(10)
                            table_cells2 = table_props2['child_artists']
                            for cell in table_cells2: cell.set_height(0.1)
                

                self.fig_options.subplots_adjust(left=0.1, bottom=0.1)    
                self.a_options.axes.get_xaxis().set_visible(False)
                self.a_options.axes.get_yaxis().set_visible(False) 
                self.a_options.margins(y=0.2)
                self.a_options.margins(x=0.2)

                
                ind = ind + 1   
                # Gets instance of subplot for onclick function
                if onclick_option == 'y':
                    self.get_ax_index.append(self.a_options) 
            
            
            figure_House.subplots_adjust(hspace=1.1)
            # Creates entirely new figure
            # If a figure has already been plotted on the window, then if this is run with 
            # create_plot == 'y' or 'Y', then the current plot on the window will double
            if create_plot == 'y':
                self.canvas_options = FigureCanvasTkAgg(figure_House, Frame_house)
                self.canvas_options.get_tk_widget().pack(fill = BOTH, expand = 1) 

    root = Tk()
    # Adjusts window size; width x height + x + y (x,y coordinates for where window will be placed on screen)
    root.geometry("1495x720+20+10")
    app = Application(master=root)
    # Calls and opens GUI
    app.mainloop()