# -*- coding: utf-8 -*-
"""
Created on Tue Dec 13 09:34:35 2016

@author: akibdpt
"""
# Imports sortal sort types, functions and rules
from sortal import (transformSet, rule, primitiveSort, point3D, coordinate, lineSegment3D, line3D, label, weight, disjunctiveSort, attributeSort, affineTransform,  vector3D)
from sympy import pi, sin, cos


# Imports API functions and variables
from api.api_create import create
from api.sort_types import sort_types
from rules.rule1_3D_RotatingSquare import rule1

# Imports GUI Library
import matplotlib as mpl
from mpl_toolkits.mplot3d import Axes3D
import numpy as np
import matplotlib.pyplot as plt
from pylab import *
import matplotlib
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import  matplotlib.pylab as plt
import matplotlib.pyplot as pltplot
from matplotlib.collections import LineCollection
matplotlib.use ("gTkAgg")
import tkinter as tk
from tkinter import ttk, Tk, Label, Button, Frame
from tkinter import *

plt.ion()

mpl.rcParams['legend.fontsize'] = 10

create = create()
LARGE_FONT= ("Verdana", 16)
NORM_FONT= ("Verdana", 10)

"""
=============
FORM CREATION
=============
'Create' methods generate and return forms based on input. As part of the API, they are imported by 
importing create from the api_create file, i.e. from api_create import create.

All sort types are initialized and stored in the file sort_types.py, inside the dictionary sort_types.
This is imported via 'from sort_types import sort_types'

Vectors are created by calling either vector2D or vector3D, as follows:
v1 = vector2D(x,y) where (x,y) are coordinates
v2 = vector3D(x,y,z) where (x,y,z) are coordinates

=======
METHODS
=======
  
Create.pointForm where a form of points is created from the inputs sort_type, and the (x,y) or (x,y,z) coordinates of an 
                 initial point which is inserted in the form. New points can be created and inserted later 
                 on into the form. This method will raise an error if the sort type passed is not based on point2D or point3D.

                 formPt = create.pointForm (sort_types['pointSort2D'], 1, 1) OR formPt = create.pointForm (sort_types['pointSort3D'], 1, 1, 1)
                 
                 To insert another point,
                 point = point2D(sort_types['pointSort2D'], 1, 0) OR point = point3D(sort_types['pointSort3D'], 1, 0, 1)
                 formPt.insert(point)
                 
Create.polygonForm where a form containing a polygon is created from an input of a list of vectors
                   These vectors serve as vertices and need to be listed in the order they will be connected
                   It is assumed that the last vector will be connected to the first vector
                   In the example, v10 will be connected to v1 to close the polygon
                    
                   This method will raise an error if the sort type passed is not based on 
                   lineSegment2D or lineSegment3D, i.e. sort_types['lineSeg2D'] OR sort_types['lineSeg3D']
                    
                   new_form3_set = [v1, v2, v3, v4, v5, v6, v7, v8, v9, v10]
                   new_form3 = create.polygonForm(sort_types['lineSeg2D'], new_form3_set)
                    
                   OR if with 3D vectors (vector3D), 
                  
                   new_form2_set = create.polygonForm(sort_types['lineSeg3D'], new_form3_set)
                    
Create.metaForm takes a list of forms as input and inserts them into a disjunctive Sort metaform. This method will raise an
                error if the sort type passed is not based on disjunctiveSort.

                forms = [formP1, formPt1]
                formM1 = create.metaForm(sort_types['disjunctive2D'], forms)
"""    

if __name__ == "__main__":
    v1 = vector3D(0, 0, 0)
    v2 = vector3D(5, 0, 0)
    v3 = vector3D(5, 3, 4)
    v4 = vector3D(0, 3, 4)

    formP1 = create.pointForm(sort_types['pointSort3D'], 4, 3, 4)

    formL1_set = [v1, v2, v3, v4]
    formL1 = create.polygonForm(sort_types['lineSeg3D'], formL1_set)

    forms = [formL1, formP1]
    formM1 = create.metaForm(sort_types['disjunctive3D'], forms)
    
    r = rule1()

    class Application(Frame):
    
        print ("<<<< GENERATING APPLICATION WINDOW >>>>")
        fig_main = Figure(figsize = (6, 6), dpi = 150) # Figure contains 'Current Drawing'
        fig_rules = Figure(figsize = (6, 8), dpi = 100) # Figure contains 'Rules'
        count = 1
        get_match = []
        
        def __init__(self, master = None):
            Frame.__init__(self, master)
            self.grid()
            self.master.title("SGI Test Case: Rotating 3D Squares")
    
            # Generates tkinter window grid with 6 rows and 6 columns
            for r in range(6):
                self.master.rowconfigure(r, weight=1)    
            for c in range(6):
                self.master.columnconfigure(c, weight=1)
    
            # Holds 'Current Drawing' figure (fig_main)
            self.Frame1 = Frame(master, bg="gray")
            self.Frame1.grid(row = 0, column = 0, rowspan = 6, columnspan = 3, sticky = W+E+N+S) 
            self.Frame1_label = Label (master = self.Frame1, text = "Current Drawing, Iteration {}".format(self.count), background = "gray", font = LARGE_FONT, anchor = "center")
            self.Frame1_label.pack()
            self.Frame2_label = Label(master = self.Frame1, text = "Match(es) is/are highlighted in purple when the cursor hovers over the plot. \nDouble click on the plot to apply the rule. \nTo rotate the plot, click on the plot and drag the cursor.", background = 'orange', font = NORM_FONT, anchor = "center")
            self.Frame2_label.pack(fill = BOTH, expand = 1)
            
            # Holds 'Rules' figure (fig_rules)
            # LHS and RHS figures are side by side for each rule
            self.Frame3 = Frame(master, bg='gray')
            self.Frame3.grid(row = 0, column = 3, rowspan = 6, columnspan = 3, sticky = W+E+N+S)
            self.Frame3_label = Label (master = self.Frame3, text ="Rule(s)", background = 'gray', font = LARGE_FONT, anchor = "center")
            self.Frame3_label.pack()
            
            self.menu() # Initializes menu
            self.set_figure() # Embeds matplotlib figures 'fig_main' and 'fig_rules' in tkinter window
            self.viz(form = formM1.elements[1].elements) # Plots initial form onto 'fig_main', 'a_main'
            self.rules_viz() # Plots rules' LHS and RHs onto 'fig_rules', 'a_rules'
            self.connect() # Activates mouseover, mouseoff and mouseclick functions
            
        """
        =====================
        API: RULE APPLICATION
        =====================
        
        'generate' applies the rule to formM1.
        
        'generate_matches' detects the matches in the current formM1, and returns the match form to be plotted.
        """
        def generate(self):
            r.apply(formM1)
            
        def generate_matches(self):
            r.detect(formM1)
            return r.detection.matches()[0].elements[1].elements, r.detection.matches()[0].elements[0].elements
        """
        ========================
        GRAPHICAL USER INTERFACE
        ========================
        """
        # Sets up main figure axis
        def call_axes(self):
            self.a_main = self.fig_main.add_subplot(111, projection = '3d')
            self.a_main.mouse_init()  
            self.a_main.set_xticks([])                               
            self.a_main.set_yticks([])                               
            self.a_main.set_zticks([])
            self.a_main.autoscale(enable=True, axis='both', tight=True)

        # Embeds main figure and rules figure in tkinter window
        def set_figure(self):
            self.canvas_main = FigureCanvasTkAgg(self.fig_main, self.Frame1)
            self.canvas_main.get_tk_widget().pack(fill = BOTH, expand = 1) 
            
            self.canvas_rules = FigureCanvasTkAgg(self.fig_rules, self.Frame3)
            self.canvas_rules.get_tk_widget().pack(fill = BOTH, expand = 1) 
    
        # Plots figures from initial form to after every rule application
        def viz(self, form, color = 'orange', viz_type = 'main', sup = 'n'):  
            self.get_match = []
            
            # Initalizes subplot in 'fig_main' and adjusts plot settings
            if sup == 'n':
                self.call_axes() 
            
            # Prepares data for plotting
            lineSeg_x, lineSeg_y, lineSeg_z = self.data_prep(form, viz_type) 
            
            # Plots data
            for i in range(len(lineSeg_x)):
                a = self.a_main.plot(lineSeg_x[i], lineSeg_y[i], zs = lineSeg_z[i], zdir = 'z', color = color, linewidth = 1.3, picker = 2)
                # Appends lines relating to match form if superposition variable (sup) is 'y' for mouse over and mouse off events
                if sup == 'y':
                    self.get_match.append(a)

            # Draws point
            self.draw_point(formM1.elements[0].elements)
        
        # Draws point/s onto plane/s
        def draw_point(self, form, color = 'black'):
            # Adjusts point size depending on iteration, as squares becomes smaller
            if self.count >= 4:
                marker_size = 9
            else:
                marker_size = 15
            # Plots point, keeps track of point instance as a variable for visualization in mouse over and mouse off events
            self.pt = self.a_main.scatter(float(form[0].value.x), float(form[0].value.y), float(form[0].value.z), 'o', s = marker_size, c = color, edgecolor = color)  
         
        # Prepares plane vertices for plotting as list of coordinates (x,y,z)
        def data_prep(self, line_set, viz_type):
            lineSeg_x = [[0 for x in range (2)] for y in range(len(line_set))]
            lineSeg_y = [[0 for x in range (2)] for y in range(len(line_set))]
            lineSeg_z = [[0 for x in range (2)] for y in range(len(line_set))]
            for i in range(len(line_set)):
                for j in range(2):
                    if viz_type == 'main':
                        if j == 0 :
                            lineSeg_x[i][j] = float(line_set[i].start.x.value)
                            lineSeg_y[i][j] = float(line_set[i].start.y.value)
                            lineSeg_z[i][j] = float(line_set[i].start.z.value)
                        else: 
                            lineSeg_x[i][j] = float(line_set[i].end.x.value)
                            lineSeg_y[i][j] = float(line_set[i].end.y.value)
                            lineSeg_z[i][j] = float(line_set[i].end.z.value)
                    elif viz_type == 'match':
                        if j == 0 :
                            lineSeg_x[i][j] = float(line_set[i].start.x)
                            lineSeg_y[i][j] = float(line_set[i].start.y)
                            lineSeg_z[i][j] = float(line_set[i].start.z)
                        else: 
                            lineSeg_x[i][j] = float(line_set[i].end.x)
                            lineSeg_y[i][j] = float(line_set[i].end.y)
                            lineSeg_z[i][j] = float(line_set[i].end.z)
            return lineSeg_x, lineSeg_y, lineSeg_z
        
        # Visualizes rule's LHS and RHS in Frame 2
        def rules_viz(self):
            rules = [r.lhs, r.rhs]
            for i in range (len(rules)):
                # Sets up each subplot for each side
                self.a_rules = self.fig_rules.add_subplot(2, 1, i+1, projection = '3d')
                # Activates rotation of subplot 
                self.a_rules.mouse_init()  
                
                # Removes axis ticks and labels, formats subplot
                self.a_rules.set_xticks([])                               
                self.a_rules.set_yticks([])                               
                self.a_rules.set_zticks([])
                self.a_rules.axes.set_aspect(0.9)
                self.a_rules.autoscale(enable=True, axis='both', tight=True)
                
                # Sets title of subplot
                if i%2 == 0:
                    self.a_rules.set_title('Rule 1, LHS', fontsize = 9)
                else:
                   self.a_rules.set_title('Rule 1, RHS', fontsize = 9)
                       
                # Plots point inside rule side
                point = rules[i].elements[0].elements[0]
                self.a_rules.scatter(float(point.value.x), float(point.value.y), float(point.value.z), 'o', s = 15, c = 'orange',  edgecolor ='orange')  
                
                # Plots lines inside rule side
                lineSeg_x, lineSeg_y, lineSeg_z = self.data_prep(rules[i].elements[1].elements, viz_type = 'match')
                for i in range(len(lineSeg_x)):
                    a = self.a_rules.plot(lineSeg_x[i], lineSeg_y[i], zs = lineSeg_z[i], zdir = 'z', color = 'blue', linewidth = 1.3, picker = 10)    
                
        # Connects mouse event functions to the 'Current Drawing' figure (fig_main)
        def connect(self):
            self.cidpress = self.fig_main.canvas.mpl_connect('button_press_event', self.on_press)
            self.cidmotion_on = self.fig_main.canvas.mpl_connect('axes_enter_event', self.on_motion)
            self.cidmotion_off = self.fig_main.canvas.mpl_connect('axes_leave_event', self.off_motion)
         
        # Calls functions to draw result of rule application 
        def on_press(self, event):     
            if event.inaxes == self.a_main and event.dblclick:
                self.count += 1
                self.generate() # Applies rule to form
                self.fig_main.clf() # Clears drawing
                self.viz(form = formM1.elements[1].elements) # Replots 'Current Drawing' after rule has been applied
                self.fig_main.canvas.draw_idle() # Draws figure
                self.Frame1_label.configure(text="Current Drawing, Iteration {}".format(self.count)) # Changes Frame 1 label
    
        # Identifies which subplot the mouse is hovering over, changes edgecolor of plane match to orange
        def on_motion(self, event):
            if event.inaxes == self.a_main:
                lines, point = self.generate_matches() # Generates match and separates them into line segments and point
                self.viz(form = lines, color = "purple", viz_type = "match", sup = 'y') # Draws line segment matches
                self.pt.remove() # Removes previous point
                self.draw_point(point, color = 'purple') # Draws point match in purple
                self.fig_main.canvas.draw_idle() # Redraws figure with new additions
    
        # Generates error box if mouse click is not within area of match
        def error_box(self):
            error_box = Tk()
            error_box.title("Input Error")
            Label(error_box, text = "Please click directly on the match.", padx = 5, pady = 15).pack()
            Button(error_box, text= "OK", padx = 5, pady = 5, command= error_box.destroy).pack()  
            error_box.update()
        
        # Removes orange color from edges of plane match and point match when  mouse cursor is not hovering over axis
        def off_motion(self, event):
            if event.inaxes == self.a_main:
                for i in self.get_match:
                    i[0].remove()
                lines, point = self.generate_matches() # Recolors match lines orange
                self.viz(form = lines, color = "orange", viz_type = "match", sup = 'y')
                self.pt.remove() # Removes purple point
                self.draw_point(point, color = 'black') # Recolors point orange
                self.fig_main.canvas.draw_idle() # Redraws figure
            
        # Initializes menu
        def menu(self):
            menubar = Menu (root)  
            menubar.add_command(label="Save Current Drawing", command=self.save) # Saves PDF of 'Current Drawing' only
            menubar.add_command(label="Exit", command=root.destroy) # Closes window
            root.config(menu=menubar)
            
        # Saves file of 'Current Drawing'
        # File type may be changed by change the .pdf extension to e.g.: .jpg, .png 
        def save(self):
            self.fig_main.savefig('currentdrawing.pdf', bbox_inches='tight')
 
    root = Tk()
    # Adjusts window size; width x height + x + y (x,y coordinates for where window will be placed on screen)
    root.geometry("1100x600+30+20")
    app = Application(master=root)
    # Calls and opens GUI
    app.mainloop()