# -*- coding: utf-8 -*-
"""
Created on Fri Feb  9 11:57:16 2018

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .bezierCurve import bezierCurve2D, bezierCurve3D
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..io.parseReader import ParsingError

class bezierCurveP2D(bezierCurve2D):
    
    type = 'ind-bezierCurveP2D'
    
    def __init__(self, sort, p1, p2, p3, isFunction = False, *args, **kwargs):
        super(bezierCurveP2D, self).__init__(sort, p1, p2, p3, isFunction)
    
    def codescriptor(self):
        return self
        
    def parametricNode(self):
        curve = bezierCurveP2D(self.sort, self.a, self.b, self.c, True)
        curve.assoc = self
        return curve
        

    def transformInd(self, transformSet):
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is None:
            return self.duplicateInd()
        p1 = transf.apply(self.start)
        p2 = transf.apply(self.controlPoint)
        p3 = transf.apply(self.end)
        return bezierCurveP2D(self.sort, p1, p2, p3)
        
    def checkTransform(self, transform, compareForm):
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        p0 = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        p1 = vector2D(x,y)
        
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        p2 = vector2D(x,y)
            
        ind = bezierCurveP2D(sort, p0, p1, p2)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line segment')
            
        reader.newToken() #next token
        
        return ind

class bezierCurveP3D(bezierCurve3D):
    
    type = 'ind-bezierCurveP3D'
    
    def __init__(self, sort, p1, p2, p3, *args, **kwargs):
        super(bezierCurveP3D, self).__init__(sort, p1, p2, p3)
    
    def codescriptor(self):
        return self
        
    def parametricNode(self):
        curve = bezierCurveP3D(self.sort, self.position(self.start), self.position(self.controlPoint), self.position(self.end))
        curve.assoc = self
        return curve
        

    def transformInd(self, transformSet):
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is None:
            return self.duplicateInd()
        p1 = transf.apply(self.position(self.start))
        p2 = transf.apply(self.position(self.controlPoint))
        p3 = transf.apply(self.position(self.end))
        return bezierCurveP3D(self.sort, p1, p2, p3)
        
    def checkTransform(self, transform, compareForm):
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y,z = vector3D.readVector(reader)
        p0 = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        p1 = vector3D(x,y,z)
        
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        p2 = vector3D(x,y,z)
            
        ind = bezierCurveP3D(sort, p0, p1, p2)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line segment')
            
        reader.newToken() #next token
        
        return ind

