# -*- coding: utf-8 -*-
"""
Created on Fri Jan 26 12:17:52 2018

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .circle import circle2D, circle3D
from ..struct.coordinate import c2
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..io.parseReader import ParsingError


class circleP2D(circle2D):
    
    type = 'ind-circleP2D'
    
    def __init__(self, sort, center, start, end, *args, **kwargs):
        super(circleP2D, self).__init__(sort, center, start, end)
        
    def equals(self, other):
        if self.center == other.center:
            return self.radius == other.radius
        return False        
    
    def codescriptor(self):
        return self
        
    def parametricNode(self):
        circle = circleP2D(self.sort, self.center, self.start, self.end)
        circle.assoc = self
        return circle
        

    def transformInd(self, transformSet):
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is None:
            return self.duplicateInd()
        center = transf.apply(self.center)
        newPts = []
        for i in transf.vects:
            if self.containsPoint(i[0]):
                newPts.append(i[1])
        if len(newPts) < 1:
            return self
        return circleP2D(self.sort, center, newPts[0], newPts[-1])
        
    def checkTransform(self, transform, compareForm):
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y = vector2D.readVector(reader)
        center = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        start = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        end = vector2D(x,y)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
            
        reader.newToken() #next token
        
        return circleP2D(sort, center, start, end)



class circleP3D(circle3D):
    
    type = 'ind-circleP3D'
    
    def __init__(self, sort, center, start, end, *args, **kwargs):
        super(circleP3D, self).__init__(sort, center, start, end)
        
    def equals(self, other):
        if self.center == other.center:
            return self.radius == other.radius
        return False        
    
    def codescriptor(self):
        return self
        
    def parametricNode(self):
        circle = circleP3D(self.sort, self.position(self.center), self.position(self.start), self.position(self.end))
        circle.assoc = self
        return circle
        

    def transformInd(self, transformSet):
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is None:
            return self.duplicateInd()
        center = transf.apply(self.position(self.center))
        newPts = []
        for i in transf.vects:
            if self.containsPoint(i[0]):
                newPts.append(i[1])
        if len(newPts) < 2:
            return self
        return circleP3D(self.sort, center, newPts[0], newPts[1])
        
        
    def checkTransform(self, transform, compareForm):
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y,z = vector3D.readVector(reader)
        center = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        start = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        end = vector3D(x,y,z)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token
        
        return circleP3D(sort, center, start, end)