# -*- coding: utf-8 -*-
"""
Created on Fri Jun 23 10:30:22 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .line import line2D, line3D
from .lineSegment import lineSegment2D, lineSegment3D
from .plane import plane2D, plane3D
from ..struct.vector import vector2D, vector3D
from ..io.parseReader import ParsingError
from ..transform.graphTransform import graphTransform, graphTransform3D


class lineSegmentP2D(lineSegment2D):
    
    type = 'ind-lineSegmentP2D'
    
    def __init__(self, sort, start = vector2D(0,0), end = vector2D(0,0), *args, **kwargs):
        super(lineSegmentP2D, self).__init__(sort, start, end)
        
    def equals(self, other):
        return (self.start == other.start) and (self.end == other.end)
        
    
    def codescriptor(self):
        return self
        
    def parametricNode(self):
        line = line2D(self.sort, self.start, self.end)
        line.assoc = self
        return line
        

    def transformInd(self, transformSet):
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is None:
            return self.duplicateInd()
        start = transf.apply(self.start)
        end = transf.apply(self.end)
        return lineSegmentP2D(self.sort, start, end)
        
    def checkTransform(self, transform, compareForm):
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        start = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        end = vector2D(x,y)
            
        ind = lineSegmentP2D(sort, start, end)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line segment')
            
        reader.newToken() #next token
        
        return ind
    
    
class lineSegmentP3D(lineSegment3D):
    
    type = 'ind-lineSegmentP3D'
    
    def __init__(self, sort, start = vector3D(0,0,0), end = vector3D(0,0,0), *args, **kwargs):
        super(lineSegmentP3D, self).__init__(sort, start, end)
        
    def equals(self, other):
        return (self.start == other.start) and (self.end == other.end)
    
    def codescriptor(self):
        return self
        
    def parametricNode(self):
        line = line3D(self.sort.getSort('__dummyLine3D__'), self.start, self.end)
        line.assoc = self
        return line
        

    def transformInd(self, transformSet):
        transf = transformSet.retrieveTransform(graphTransform3D)
        #print(transformSet.map)
        if transf is None:
            return self.duplicateInd()
        start = transf.apply(self.start)
        end = transf.apply(self.end)
        return lineSegmentP3D(self.sort, start, end)
        
    def checkTransform(self, transform, compareForm):
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y,z = vector3D.readVector(reader)
        start = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        end = vector3D(x,y,z)
            
        ind = lineSegmentP3D(sort, start, end)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line segment')
            
        reader.newToken() #next token
        
        return ind
    
    
    