#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

from sortal import form, sort

"""
=============
FUNCTION INFO
=============
'part_of' checks if a subshape is part of a possibly larger shape

INPUT
subshape: List of Rhino GUIDs or name of subshape
shape: List of Rhino GUIDs or name of shape

OUTPUT
Boolean value True/False; True, if subshape are part of the shape; False, if subshape is not part of the shape

WARNINGS / ERRORS
KeyError: If either shape or subshape name is not available in the form register
          OR if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
TypeError: If either shape or subshape input is not a text string

Data Flow: agnostic -> sortal -> Boolean Output
"""
   
def part_of(subshape, shape):
    if 'rhino_shapes_c' in sort.register:
        # Checks if the input for shape is correct
        # Retrieves shape object from form register
        if type(shape) == type(''):
            if shape in form.formRegister:
                shapeSortal = form.formRegister[shape]
            else:
                raise KeyError('Shape with name ' + shape +' is not available in the form register')
        else:
            raise TypeError('Shape input is not a text string')
            
        # Checks if the input for subshape is correct
        # Retrieves subshape object from form register
        if type(subshape) == type('a'):
            if subshape in form.formRegister:
                subshapeSortal = form.formRegister[subshape]
            else:
                raise KeyError('Shape with name ' + subshape +' is not available in the form register')
        else:
            raise TypeError('Subshape input is not a text string')
                   
        return subshapeSortal.partOf(shapeSortal)
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                
